#include <sys/types.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include "fw.h"

struct stat		statbuf;
struct FW_header	*FW;
char			*buffer;
char			prefix[256] = "firmware";
char			file[256];

void dump_section( char *fn, char *buffer, u_int32_t start, u_int32_t size )
{
	int		fd, n;
	u_int32_t	i;

	if( strlen( fn ) == 0 )
		return;

	fprintf( stdout, "Dump Section %s, 0x%08x, %10d\n", fn, start, size );

	if ( (fd = open( fn, O_WRONLY|O_CREAT, 0666 )) < 0 )
	{
		fprintf( stderr, "Open Failed %s: %s\n",
					fn, strerror( errno ) );
		return;
	}

	buffer += start;

	for( i = 0; i < size; )
	{
		if ( (n = size - i) > 1024 )
		{
			n = 1024;
		}

		n = write( fd, buffer + i, n );

		if ( n < 0 )
		{
			fprintf( stderr, "Write Error %s: %s\n",
					fn, strerror( errno ) );
			return;
		}
		i += n;
	}
}

int main( int argc, char *argv[] )
{
	int		fd, r, i;
	char		*p, opt;
	u_int32_t	c;
	int		hflag = 0, qflag = 0, pflag = 0;
	char		*exec_name = 0;
	char		*inputname = 0;
	char		*kernelname = 0;
	char		*ramdiskname = 0;
	char		*extraname = 0;

	exec_name = argv[0];

	while ((opt = getopt(argc, argv, "hqk:r:e:p")) > 0) {
		switch( opt ) {
		case 'h':	/* dump just the header */
			hflag = 1;
			break;
		case 'q':	/* quiet, don't print header */
			qflag = 1;
			break;
		case 'k':	/* kernel image file name */
			kernelname = optarg;
			break;
		case 'r':	/* ramdisk image file name */
			ramdiskname = optarg;
			break;
		case 'e':	/* extra section file name */
			extraname = optarg;
			break;
		case 'p':
			pflag = 1;
			break;
		default:
			fprintf(stderr,
				"Usage: %s [-hqp] [-k kname] [-r rname] "
				"[-e ename] <inputfilename>\n",
			exec_name );
			exit( 1 );
		}
	}

	if( !qflag )
		fprintf( stdout, "%s version %d.%d\n", exec_name,
				FW_VERSION, FW_REVISION );

	inputname = argv[optind];
	if( (fd = open( inputname, O_RDONLY ) ) < 0 )
	{
		fprintf( stderr, "%s: Can't open %s: %s\n", exec_name,
				inputname, strerror( errno ) );
		exit( -1 );
	}

	if( fstat( fd, &statbuf ) < 0 )
	{
		fprintf( stderr, "%s: Can't get file size, %s: %s\n",
				exec_name, inputname, strerror( errno ) );
		exit( -1 );
	}

	if( ( buffer = malloc( statbuf.st_size ) ) == NULL )
	{
		fprintf( stderr, "%s: Can't allocate memory %s\n",
				exec_name, strerror( errno ) );
		exit( -1 );
	}

	for( p = buffer; (r = read(fd, p, 1024)) > 0; p += r )
		;
	if( r < 0 )
	{
		fprintf( stderr, "%s: Read error %s: %s\n",
				exec_name, inputname, strerror( errno ) );
		exit( -1 );
	}

	FW = (struct FW_header *)buffer;

	if ( !qflag )
	{
		fprintf( stdout, "Section 1 %8d, %8d\n", ntohl( FW->start_1 ),
			ntohl( FW->size_1 ) );
		fprintf( stdout, "Section 2 %8d, %8d\n", ntohl( FW->start_2 ),
			ntohl( FW->size_2 ) );
		fprintf( stdout, "Section 3 %8d, %8d\n", ntohl( FW->start_3 ),
			ntohl( FW->size_3 ) );
		fprintf( stdout, "\nMagic      " );
		for( i = 0; i < 12; i++ )
		{
			fprintf( stdout, " 0x%02x", FW->magic[i] );
		}
		fprintf( stdout, "\n" );
		fprintf( stdout,
			"Product ID: 0x%02x, Custom ID: "
			"0x%02x, Model ID: 0x%02x\n\n",
			FW->productId, FW->customId, FW->modelId );

	}
	c = cksum_section( buffer, ntohl( FW->start_1 ), ntohl( FW->size_1 ) );

	if ( FW->checksum_1 !=  c )
	{
		fprintf( stderr, "Invalid checksum section 1\n" );
		exit( -1 );
	}

	c = cksum_section( buffer, ntohl( FW->start_2 ), ntohl( FW->size_2 ) );

	if ( FW->checksum_2 !=  c )
	{
		fprintf( stderr, "Invalid checksum section 2\n" );
		exit( -1 );
	}

	c = cksum_section( buffer, ntohl( FW->start_3 ), ntohl( FW->size_3 ) );

	if ( FW->checksum_3 !=  c )
	{
		fprintf( stderr, "Invalid checksum section 3\n" );
		exit( -1 );
	}

	if( hflag )
		exit( 0 );

	if( !kernelname )
		snprintf( file, sizeof( file ), "%s.sec1", prefix );
	else
		strncpy( file, kernelname, sizeof( file ) );
	dump_section( file, buffer, ntohl( FW->start_1 ), ntohl( FW->size_1 ) );

	if( !ramdiskname )
		snprintf( file, sizeof( file ), "%s.sec2", prefix );
	else
		strncpy( file, ramdiskname, sizeof( file ) );
	dump_section( file, buffer, ntohl( FW->start_2 ), ntohl( FW->size_2 ) );

	if( !extraname )
		snprintf( file, sizeof( file ), "%s.sec3", prefix );
	else
		strncpy( file, extraname, sizeof( file ) );
	dump_section( file, buffer, ntohl( FW->start_3 ), ntohl( FW->size_3 ) );

	exit( 0 );
}
